@@ -4,22 +4,17 @@
  *
  * Copyright (C) 2003 NVIDIA Corporation
  *
- * This program is free software; you can redistribute it and/or
- * modify it under the terms of the GNU General Public License as
- * published by the Free Software Foundation; either version 2 of the
- * License, or (at your option) any later version.
+ * This program is free software; you can redistribute it and/or modify it
+ * under the terms and conditions of the GNU General Public License,
+ * version 2, as published by the Free Software Foundation.
  * 
- * This program is distributed in the hope that it will be useful, but
- * WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
- * General Public License for more details.
+ * This program is distributed in the hope that it will be useful, but WITHOUT
+ * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
+ * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
+ * more details.
  * 
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the:
- *
- *      Free Software Foundation, Inc.
- *      59 Temple Place - Suite 330
- *      Boston, MA 02111-1307, USA
+ * along with this program; if not, see <http://www.gnu.org/licenses>.
  *
  *
  * command_list.c - this source file contains functions for building
@@ -39,6 +34,7 @@
 #include <stdio.h>
 #include <stdarg.h>
 #include <errno.h>
+#include <utime.h>
 
 #include "nvidia-installer.h"
 #include "command-list.h"
@@ -47,27 +43,16 @@
 #include "misc.h"
 #include "files.h"
 #include "kernel.h"
+#include "manifest.h"
+#include "conflicting-kernel-modules.h"
 
 
 static void free_file_list(FileList* l);
 
-static void find_conflicting_xfree86_libraries(Options *,
-                                               const char *,
-                                               FileList *);
+static void find_conflicting_kernel_modules(Options *op, FileList *l);
 
-static void find_conflicting_xfree86_libraries_fullpath(Options *op,
-                                                        const char *,
-                                                        FileList *l);
-
-static void find_conflicting_opengl_libraries(Options *,
-                                              const char *,
-                                              FileList *);
-
-static void find_conflicting_kernel_modules(Options *op,
-                                            Package *p,
-                                            FileList *l);
-
-static void find_existing_files(Package *p, FileList *l, uint64_t);
+static void find_existing_files(Package *p, FileList *l,
+                                PackageEntryFileTypeList *file_type_list);
 
 static void condense_file_list(Package *p, FileList *l);
 
@@ -77,6 +62,123 @@ static void add_file_to_list(const char*, const char*, FileList*);
 
 static void append_to_rpm_file_list(Options *op, Command *c);
 
+static ConflictingFileInfo *build_conflicting_file_list(Options *op, Package *p);
+static void get_conflicting_file_info(const char *file, ConflictingFileInfo *cfi);
+
+/*
+ * find_conflicting_files() optionally takes an array of NoRecursionDirectory
+ * entries to indicate which directories should not be recursively searched.
+ */
+
+typedef struct {
+    int level;   /* max search depth: set to negative for unrestricted depth */
+    char *name;  /* name to find: NULL to end the list */
+} NoRecursionDirectory;
+
+static void find_conflicting_files(Options *op,
+                                   char *path,
+                                   ConflictingFileInfo *files,
+                                   FileList *l,
+                                   const NoRecursionDirectory *skipdirs);
+
+
+/*
+ * Check if a path already exists in the path list, or is a subdirectory of
+ * a path that exists in the path list, or is a symlink to or symlink target
+ * of a directory that exists in the path list.
+ */
+static int path_already_exists(char ***paths, int count, const char *path)
+{
+    int i;
+
+    for (i = 0; i < count; i++) {
+        int is_subdir = FALSE;
+
+        is_subdirectory((*paths)[i], path, &is_subdir);
+
+        if (is_subdir) {
+            return TRUE;
+        }
+    }
+
+    return FALSE;
+}
+
+/*
+ * Add a new path to the list of paths to search, provided that it exists
+ * and is not redundant.
+ * XXX we only check to see if the new directory is a subdirectory of any
+ * existing directory, and not the other way around.
+ */
+static void add_search_path(char ***paths, int *count, const char *path)
+{
+    if (directory_exists(path) && !path_already_exists(paths, *count, path)) {
+        *paths = nvrealloc(*paths, sizeof(char *) * (*count + 1));
+        (*paths)[*count] = nvstrdup(path);
+        (*count)++;
+    }
+}
+
+/*
+ * Given a path, add the subdirectories "/lib", "/lib32", and "/lib64" to the
+ * list of paths to search.
+ */
+static void add_search_paths(char ***paths, int *count, const char *pathbase)
+{
+    int i;
+    const char *subdirs[] = {
+        "/lib",
+        "/lib32",
+        "/lib64",
+    };
+
+    for (i = 0; i < ARRAY_LEN(subdirs); i++) {
+        char *path = nvstrcat(pathbase, subdirs[i], NULL);
+        add_search_path(paths, count, path);
+        nvfree(path);
+    }
+}
+
+/*
+ * Build the list of paths under which to search for conflicting files.
+ * Returns the number of paths added to the search list.
+ */
+static int get_conflicting_search_paths(const Options *op, char ***paths)
+{
+    int ret = 0;
+
+    *paths = NULL;
+
+    add_search_paths(paths, &ret, DEFAULT_X_PREFIX);
+    add_search_paths(paths, &ret, XORG7_DEFAULT_X_PREFIX);
+    add_search_paths(paths, &ret, op->x_prefix);
+    add_search_paths(paths, &ret, DEFAULT_OPENGL_PREFIX);
+    add_search_paths(paths, &ret, op->opengl_prefix);
+    add_search_path(paths, &ret, op->x_module_path);
+    add_search_path(paths, &ret, op->x_library_path);
+
+#if defined(NV_X86_64)
+    if (op->compat32_chroot != NULL) {
+        int i;
+        char *subdirs[] = {
+            DEFAULT_X_PREFIX,
+            op->x_prefix,
+            DEFAULT_OPENGL_PREFIX,
+            op->opengl_prefix,
+            op->compat32_prefix,
+        };
+
+        for (i = 0; i < ARRAY_LEN(subdirs); i++) {
+            char *path = nvstrcat(op->compat32_chroot, "/", subdirs[i], NULL);
+            add_search_paths(paths, &ret, path);
+            nvfree(path);
+        }
+    }
+#endif
+
+    return ret;
+}
+
 
 /*
  * build_command_list() - construct a list of all the things to do
@@ -85,7 +187,7 @@ static void append_to_rpm_file_list(Options *op, Command *c);
  *   - backing and removing up conflicting files
  *   - installing all the installable files
  *   - create any needed symbolic links
- *   - executing any nessary commands
+ *   - executing any necessary commands
  *   - running `ldconfig` and `depmod -aq`
  *
  * If we are only installing the kernel module, then we trim back all
@@ -97,101 +199,34 @@ CommandList *build_command_list(Options *op, Package *p)
     FileList *l;
     CommandList *c;
     int i, cmd;
-    uint64_t installable_files;
+    PackageEntryFileTypeList installable_files;
+    PackageEntryFileTypeList tmp_installable_files;
     char *tmp;
 
-    installable_files = get_installable_file_mask(op);
-    
+    get_installable_file_type_list(op, &installable_files);
+
     l = (FileList *) nvalloc(sizeof(FileList));
     c = (CommandList *) nvalloc(sizeof(CommandList));
 
-    /* find any possibly conflicting libraries and/or modules */
-    
-    if (!op->kernel_module_only) {
-        /*
-         * Note that searching the various paths may produce duplicate
-         * entries for conflicting files; this is OK because we will take
-         * care of these duplicates in condense_file_list().
-         */
+    /* find any possibly conflicting modules and/or libraries */
 
-        ui_status_begin(op, "Searching for conflicting X files:", "Searching");
-
-        ui_status_update(op, 0.16f, DEFAULT_X_PREFIX);
-        find_conflicting_xfree86_libraries(op, DEFAULT_X_PREFIX, l);
-        ui_status_update(op, 0.32f, XORG7_DEFAULT_X_PREFIX);
-        find_conflicting_xfree86_libraries(op, XORG7_DEFAULT_X_PREFIX, l);
-        ui_status_update(op, 0.48f, op->x_prefix);
-        find_conflicting_xfree86_libraries(op, op->x_prefix, l);
-
-        ui_status_update(op, 0.64f, op->x_module_path);
-        find_conflicting_xfree86_libraries_fullpath(op, op->x_module_path, l);
-        ui_status_update(op, 0.80f, op->x_library_path);
-        find_conflicting_xfree86_libraries_fullpath(op, op->x_library_path, l);
-
-        ui_status_end(op, "done.");
-
-        ui_status_begin(op, "Searching for conflicting OpenGL files:", "Searching");
-
-        ui_status_update(op, 0.20f, DEFAULT_X_PREFIX);
-        find_conflicting_opengl_libraries(op, DEFAULT_X_PREFIX, l);
-        ui_status_update(op, 0.40f, op->x_prefix);
-        find_conflicting_opengl_libraries(op, op->x_prefix, l);
-        ui_status_update(op, 0.60f, DEFAULT_OPENGL_PREFIX);
-        find_conflicting_opengl_libraries(op, DEFAULT_OPENGL_PREFIX, l);
-        ui_status_update(op, 0.80f, op->opengl_prefix);
-        find_conflicting_opengl_libraries(op, op->opengl_prefix, l);
-
-        ui_status_end(op, "done.");
-
-#if defined(NV_X86_64)
-        if (op->compat32_chroot != NULL) {
-            char *prefix;
-
-            ui_status_begin(op, "Searching for conflicting compat32 files:", "Searching");
-
-            prefix = nvstrcat(op->compat32_chroot, DEFAULT_X_PREFIX, NULL);
-            ui_status_update(op, 0.20f, prefix);
-            find_conflicting_opengl_libraries(op, prefix, l);
-            nvfree(prefix);
-
-            prefix = nvstrcat(op->compat32_chroot, op->x_prefix, NULL);
-            ui_status_update(op, 0.40f, prefix);
-            find_conflicting_opengl_libraries(op, prefix, l);
-            nvfree(prefix);
-
-            prefix = nvstrcat(op->compat32_chroot, DEFAULT_OPENGL_PREFIX, NULL);
-            ui_status_update(op, 0.60f, prefix);
-            find_conflicting_opengl_libraries(op, prefix, l);
-            nvfree(prefix);
-
-            prefix = nvstrcat(op->compat32_chroot, op->compat32_prefix, NULL);
-            ui_status_update(op, 0.80f, prefix);
-            find_conflicting_opengl_libraries(op, prefix, l);
-            nvfree(prefix);
-
-            ui_status_end(op, "done.");
-        }
-#endif /* NV_X86_64 */
+    if (!op->no_kernel_modules) {
+        find_conflicting_kernel_modules(op, l);
     }
-    
-    if (!op->no_kernel_module)
-        find_conflicting_kernel_modules(op, p, l);
-    
-    /*
-     * find any existing files that clash with what we're going to
-     * install
-     * DON'T include files of type FILE_TYPE_NEWSYM
-     */
-    
-    find_existing_files(p, l, installable_files | FILE_TYPE_SYMLINK);
-    
-    /* condense the file list */
 
-    condense_file_list(p, l);
-    
-    /* check the conflicting file list for any installed files */
+    /* check the conflicting file list for any installed kernel modules */
+
+    if (op->kernel_modules_only) {
+        const char *kernel = get_kernel_name(op);
+
+        if (dkms_module_installed(op, p->version, kernel)) {
+            ui_error(op, "DKMS kernel modules with version %s are already "
+                     "installed for the %s kernel.", p->version, kernel);
+            free_file_list(l);
+            free_command_list(op,c);
+            return NULL;
+        }
 
-    if (op->kernel_module_only) {
         for (i = 0; i < l->num; i++) {
             if (find_installed_file(op, l->filename[i])) {
                 ui_error(op, "The file '%s' already exists as part of this "
@@ -201,8 +236,68 @@ CommandList *build_command_list(Options *op, Package *p)
                 return NULL;
             }
         }
+
+        /* XXX: If installing with --kernel-modules-only on a system that has
+         * the kernel module sources already installed, but does NOT have built
+         * kernel modules or DKMS modules, duplicate entries for the source
+         * files will be added to the backup log, leading to error messages when
+         * uninstalling the driver later leads to redundant attempts to delete
+         * the files. */
+
     }
     
+    if (!op->kernel_modules_only) {
+        char **paths;
+        int numpaths, i;
+        ConflictingFileInfo *conflicting_files;
+
+        /*
+         * stop recursing into any "nvidia-cg-toolkit"
+         * directory to prevent libGL.so.1 from being deleted
+         * (see bug 843595).
+         *
+         * Also, do not recurse into "source" or "build" directories, pretty
+         * much ever. This is because distros (for example, Fedora Core 21)
+         * have started putting links to kernel source and build trees in
+         * other locations, including /usr/lib/modules/`uname -r`/kernel.
+         * (See bug 1646361).
+         */
+        static const NoRecursionDirectory skipdirs[] = {
+            { -1, "source" },
+            { -1, "build" },
+            { -1, "nvidia-cg-toolkit" },
+            {  0, NULL }
+        };
+
+        numpaths = get_conflicting_search_paths(op, &paths);
+
+        ui_status_begin(op, "Searching for conflicting files:", "Searching");
+
+        conflicting_files = build_conflicting_file_list(op, p);
+        for (i = 0; i < numpaths; i++) {
+            ui_status_update(op, (i + 1.0f) / numpaths, "Searching: %s", paths[i]);
+            find_conflicting_files(op, paths[i], conflicting_files, l,
+                                   skipdirs);
+        }
+        nvfree(conflicting_files);
+
+        ui_status_end(op, "done.");
+    }
+    
+    /*
+     * find any existing files that clash with what we're going to
+     * install
+     */
+
+    tmp_installable_files = installable_files;
+    add_symlinks_to_file_type_list(&tmp_installable_files);
+
+    find_existing_files(p, l, &tmp_installable_files);
+    
+    /* condense the file list */
+
+    condense_file_list(p, l);
+    
     /*
      * all of the files in the conflicting file list should be backed
      * up or deleted
@@ -233,14 +328,14 @@ CommandList *build_command_list(Options *op, Package *p)
          */
         if (op->selinux_enabled &&
             (op->utils[EXECSTACK] != NULL) &&
-            (p->entries[i].flags & FILE_TYPE_SHARED_LIB)) {
+            (p->entries[i].caps.is_shared_lib)) {
             tmp = nvstrcat(op->utils[EXECSTACK], " -c ",
                            p->entries[i].dst, NULL);
         } else {
             tmp = NULL;
         }
 
-        if (p->entries[i].flags & installable_files) {
+        if (installable_files.types[p->entries[i].type]) {
             add_command(c, INSTALL_CMD,
                         p->entries[i].file,
                         p->entries[i].dst,
@@ -251,18 +346,27 @@ CommandList *build_command_list(Options *op, Package *p)
         nvfree(tmp);
 
         /*
-         * delete the temporary libGL.la and .desktop files generated
-         * based on templates earlier.
+         * For icons, update the mtime on the top-level directory of the icon
+         * theme. This triggers desktop environments to invalidate their icon
+         * cache.
+         *
+         * See https://specifications.freedesktop.org/icon-theme-spec/icon-theme-spec-latest.html#implementation_notes
+         */
+        if (p->entries[i].type == FILE_TYPE_ICON) {
+            add_command(c, TOUCH_CMD, op->icon_dir);
+        }
+
+        /*
+         * delete any temporary generated files
          */
 
-        if ((p->entries[i].flags & FILE_TYPE_LIBGL_LA) ||
-                (p->entries[i].flags & FILE_TYPE_DOT_DESKTOP)) {
+        if (p->entries[i].caps.is_temporary) {
             add_command(c, DELETE_CMD,
                         p->entries[i].file);
         }
 
         if (op->selinux_enabled &&
-            (p->entries[i].flags & FILE_TYPE_SHARED_LIB)) {
+            (p->entries[i].caps.is_shared_lib)) {
             tmp = nvstrcat(op->utils[CHCON], " -t ", op->selinux_chcon_type,
                            " ", p->entries[i].dst, NULL);
             add_command(c, RUN_CMD, tmp);
@@ -274,34 +378,12 @@ CommandList *build_command_list(Options *op, Package *p)
     /* create any needed symbolic links */
     
     for (i = 0; i < p->num_entries; i++) {
-        if ((p->entries[i].flags & FILE_TYPE_SYMLINK) ||
-            (p->entries[i].flags & FILE_TYPE_NEWSYM)) {
-            /* if it's a NEWSYM and the file already exists, don't add a command
-             * for it */
-            if (p->entries[i].flags & FILE_TYPE_NEWSYM) {
-                struct stat buf;
-                if(!stat(p->entries[i].dst, &buf) || errno != ENOENT) {
-                    ui_expert(op, "Not creating a symlink from %s to %s "
-                                  "because a file already exists at that path "
-                                  "or the path is inaccessible.",
-                                  p->entries[i].dst, p->entries[i].target);
-                    continue;
-                }
-            }
-
+        if (p->entries[i].caps.is_symlink) {
             add_command(c, SYMLINK_CMD, p->entries[i].dst,
                         p->entries[i].target);
         }
     }
     
-    /* find any commands we should run */
-
-    for (i = 0; i < p->num_entries; i++) {
-        if (p->entries[i].flags & FILE_TYPE_KERNEL_MODULE_CMD) {
-            add_command(c, RUN_CMD, p->entries[i].file, NULL, 0);
-        }
-    }
-
     /*
      * if "--no-abi-note" was requested, scan for any OpenGL
      * libraries, and run the following command on them:
@@ -310,15 +392,22 @@ CommandList *build_command_list(Options *op, Package *p)
      */
     
     if (op->no_abi_note) {
-        for (i = 0; i < p->num_entries; i++) {
-            if (p->entries[i].flags & FILE_TYPE_OPENGL_LIB) {
-                tmp = nvstrcat(op->utils[OBJCOPY],
-                               " --remove-section=.note.ABI-tag ",
-                               p->entries[i].dst,
-                               " 2> /dev/null ; true", NULL);
-                add_command(c, RUN_CMD, tmp);
-                nvfree(tmp);
+
+        if (op->utils[OBJCOPY]) {
+            for (i = 0; i < p->num_entries; i++) {
+                if (p->entries[i].type == FILE_TYPE_OPENGL_LIB) {
+                    tmp = nvstrcat(op->utils[OBJCOPY],
+                            " --remove-section=.note.ABI-tag ",
+                            p->entries[i].dst,
+                            " 2> /dev/null ; true", NULL);
+                    add_command(c, RUN_CMD, tmp);
+                    nvfree(tmp);
+                }
             }
+        } else {
+            ui_warn(op, "--no-abi-note option was specified but the system "
+                    "utility `objcopy` (package 'binutils') was not found; this "
+                    "operation will be skipped.");
         }
     }
     
@@ -333,22 +422,19 @@ CommandList *build_command_list(Options *op, Package *p)
      * <Nigel.Spowage@energis.com>
      */
     
-    if ( op->kernel_name && op->kernel_name[0] ) {
-    	tmp = nvstrcat(op->utils[DEPMOD], " -aq ", op->kernel_name, NULL);
-    } else {
-    	tmp = nvstrcat(op->utils[DEPMOD], " -aq", NULL);
+    if (!op->no_kernel_modules && !op->skip_depmod) {
+        tmp = nvstrcat(op->utils[DEPMOD], " -a ", op->kernel_name, NULL);
+        add_command(c, RUN_CMD, tmp);
+        nvfree(tmp);
     }
-    add_command(c, RUN_CMD, tmp);
-    nvfree(tmp);
-    
-    /*
-     * if on SuSE or United Linux, also do `/usr/bin/chrc.config
-     * SCRIPT_3D no`
-     */
 
-    if (((op->distro == SUSE) || (op->distro == UNITED_LINUX)) &&
-        (access("/usr/bin/chrc.config", X_OK) == 0)) {
-        add_command(c, RUN_CMD, "/usr/bin/chrc.config SCRIPT_3D no");
+    /*
+     * If systemd files were installed, run `systemctl daemon-reload`.
+     */
+    if (op->use_systemd == NV_OPTIONAL_BOOL_TRUE) {
+        tmp = nvstrcat(op->utils[SYSTEMCTL], " daemon-reload", NULL);
+        add_command(c, RUN_CMD, tmp);
+        nvfree(tmp);
     }
 
     /* free the FileList */
@@ -417,7 +503,7 @@ static inline int execute_run_command(Options *op, float percent, const char *cm
     ui_expert(op, "Executing: %s", cmd);
     ui_status_update(op, percent, "Executing: `%s` "
                      "(this may take a moment...)", cmd);
-    ret = run_command(op, cmd, &data, TRUE, 0, TRUE);
+    ret = run_command(op, cmd, &data, TRUE, NULL, TRUE);
     if (ret != 0) {
         ui_error(op, "Failed to execute `%s`: %s", cmd, data);
         ret = continue_after_error(op, "Failed to execute `%s`", cmd);
@@ -442,7 +528,7 @@ int execute_command_list(Options *op, CommandList *c,
     int i, ret;
     float percent;
 
-    ui_status_begin(op, title, msg);
+    ui_status_begin(op, title, "%s", msg);
 
     for (i = 0; i < c->num; i++) {
 
@@ -520,6 +606,16 @@ int execute_command_list(Options *op, CommandList *c,
             }
             break;
 
+        case TOUCH_CMD:
+            ui_expert(op, "Updating mtime: %s", c->cmds[i].s0);
+            ret = utime(c->cmds[i].s0, NULL);
+            if (ret == -1) {
+                ret = continue_after_error(op, "Cannot touch %s",
+                                           c->cmds[i].s0);
+                if (!ret) return FALSE;
+            }
+            break;
+
         default:
             /* XXX should never get here */
             return FALSE;
@@ -540,105 +636,6 @@ int execute_command_list(Options *op, CommandList *c,
  ***************************************************************************
  */
 
-typedef struct {
-    const char *name;
-    int len;
-
-    /*
-     * if requiredString is non-NULL, then a file must have this
-     * string in order to be considered a conflicting file; we use
-     * this to only consider "libglx.*" files conflicts if they have
-     * the string "glxModuleData".
-     */
-
-    const char *requiredString;
-} ConflictingFileInfo;
-
-static void find_conflicting_files(Options *op,
-                                   char *path,
-                                   ConflictingFileInfo *files,
-                                   FileList *l);
-
-static void find_conflicting_libraries(Options *op,
-                                       const char *prefix,
-                                       ConflictingFileInfo *libs,
-                                       FileList *l);
-
-static ConflictingFileInfo __xfree86_libs[] = {
-    { "libGLcore.",          10, /* strlen("libGLcore.") */          NULL            },
-    { "libGL.",              6,  /* strlen("libGL.") */              NULL            },
-    { "libGLwrapper.",       13, /* strlen("libGLwrapper.") */       NULL            },
-    { "libglx.",             7,  /* strlen("libglx.") */             "glxModuleData" },
-    { "libXvMCNVIDIA",       13, /* strlen("libXvMCNVIDIA") */       NULL            },
-    { "libnvidia-cfg.",      14, /* strlen("libnvidia-cfg.") */      NULL            },
-    { "nvidia_drv.",         11, /* strlen("nvidia_drv.") */         NULL            },
-    { "libcuda.",            8,  /* strlen("libcuda.") */            NULL            },
-    { "libvdpau.",           9,  /* strlen("libvdpau.") */           NULL            },
-    { "libvdpau_trace.",     15, /* strlen("libvdpau_trace.") */     NULL            },
-    { "libvdpau_nvidia.",    16, /* strlen("libvdpau_nvidia.") */    NULL            },
-    { "libnvidia-compiler.", 19, /* strlen("libnvidia-compiler.") */ NULL            },
-    { NULL,                  0,                                      NULL            }
-};
-
-/*
- * find_conflicting_xfree86_libraries() - search for conflicting
- * libraries under the XFree86 installation prefix, for all possible
- * libdirs.
- */
-
-static void find_conflicting_xfree86_libraries(Options *op,
-                                               const char *xprefix,
-                                               FileList *l)
-{
-    find_conflicting_libraries(op, xprefix, __xfree86_libs, l);
-
-} /* find_conflicting_xfree86_libraries() */
-
-
-
-/*
- * find_conflicting_xfree86_libraries_fullpath() - same as
- * find_conflicting_xfree86_libraries, but bypasses the
- * find_conflicting_libraries step, which appends "lib", "lib64", and
- * "lib32" to the path name.  Use this when you have the fullpath that
- * you want searched.
- */
-
-static void find_conflicting_xfree86_libraries_fullpath(Options *op,
-                                                        const char *path,
-                                                        FileList *l)
-{
-    find_conflicting_files(op, (char *) path, __xfree86_libs, l);
-    
-} /* find_conflicting_xfree86_libraries_fullpath() */
-
-
-
-static ConflictingFileInfo __opengl_libs[] = {
-    { "libGLcore.",          10, /* strlen("libGLcore.") */          NULL },
-    { "libGL.",              6,  /* strlen("libGL.") */              NULL },
-    { "libnvidia-tls.",      14, /* strlen("libnvidia-tls.") */      NULL },
-    { "libGLwrapper.",       13, /* strlen("libGLwrapper.") */       NULL },
-    { "libnvidia-cfg.",      14, /* strlen("libnvidia-cfg.") */      NULL },
-    { "libcuda.",            8,  /* strlen("libcuda.") */            NULL },
-    { "libnvidia-compiler.", 19, /* strlen("libnvidia-compiler.") */ NULL },
-    { NULL,                  0,                                      NULL }
-};
-
-/*
- * find_conflicting_opengl_libraries() - search for conflicting
- * libraries under the OpenGL installation prefix, for all possible
- * libdirs.
- */
-
-static void find_conflicting_opengl_libraries(Options *op,
-                                              const char *glprefix,
-                                              FileList *l)
-{
-    find_conflicting_libraries(op, glprefix, __opengl_libs, l);
-
-} /* find_conflicting_opengl_libraries() */
-
 
 
 /*
@@ -646,37 +643,50 @@ static void find_conflicting_opengl_libraries(Options *op,
  * modules under the kernel module installation prefix.
  */
 
-static void find_conflicting_kernel_modules(Options *op,
-                                            Package *p, FileList *l)
+static void find_conflicting_kernel_modules(Options *op, FileList *l)
 {
-    int i, n = 0;
-    ConflictingFileInfo files[2];
+    int i = 0;
+    ConflictingFileInfo *files;
     char *paths[3];
     char *tmp = get_kernel_name(op);
+    char **filenames;
 
-    memset(files, 0, sizeof(files));
-    files[1].name = NULL;
-    files[1].len = 0;
-    paths[0] = op->kernel_module_installation_path;
+    /* Don't descend into the "build" or "source" directories; these won't
+     * contain modules, and may be symlinks back to an actual source tree. */
+    static const NoRecursionDirectory skipdirs[] = {
+        { 1, "build" },
+        { 1, "source" },
+        { 0, NULL }
+    };
+
+    if (op->kernel_module_installation_path) {
+        paths[i++] = op->kernel_module_installation_path;
+    }
 
     if (tmp) {
-        paths[1] = nvstrcat("/lib/modules/", tmp, NULL);
-        paths[2] = NULL;
-    } else {
-        paths[1] = NULL;
+        paths[i++] = nvstrcat("/lib/modules/", tmp, NULL);
     }
-    
-    for (i = 0; paths[i]; i++) {
-        for (n = 0; p->bad_module_filenames[n]; n++) {
-            /*
-             * Recursively search for this conflicting kernel module
-             * relative to the current prefix.
-             */
-            files[0].name = p->bad_module_filenames[n];
-            files[0].len = strlen(files[0].name);
 
-            find_conflicting_files(op, paths[i], files, l);
-        }
+    paths[i] = NULL;
+
+    /* Build the list of conflicting kernel modules */
+
+    files = nvalloc((num_conflicting_kernel_modules + 1) * sizeof(files[0]));
+    filenames = nvalloc(num_conflicting_kernel_modules * sizeof(filenames[0]));
+
+    for (i = 0; i < num_conflicting_kernel_modules; i++) {
+        filenames[i] = nvstrcat(conflicting_kernel_modules[i], ".ko", NULL);
+        files[i].name = filenames[i];
+        files[i].len = strlen(filenames[i]);
+    }
+
+    for (i = 0; paths[i]; i++) {
+        /*
+         * Recursively search for the conflicting kernel modules
+         * relative to the current prefix.
+         */
+
+        find_conflicting_files(op, paths[i], files, l, skipdirs);
     }
 
     /* free any paths we nvstrcat()'d above  */
@@ -685,7 +695,14 @@ static void find_conflicting_kernel_modules(Options *op,
         nvfree(paths[i]);
     }
 
-} /* find_conflicting_kernel_modules() */
+    /* free the kernel module names */
+
+    for (i = 0; i < num_conflicting_kernel_modules; i++) {
+        nvfree(filenames[i]);
+    }
+    nvfree(filenames);
+    nvfree(files);
+}
 
 
 
@@ -695,13 +712,14 @@ static void find_conflicting_kernel_modules(Options *op,
  * FileList.
  */
 
-static void find_existing_files(Package *p, FileList *l, uint64_t flag)
+static void find_existing_files(Package *p, FileList *l,
+                                PackageEntryFileTypeList *file_type_list)
 {
     int i;
     struct stat stat_buf;
 
     for (i = 0; i < p->num_entries; i++) {
-        if (p->entries[i].flags & flag) {
+        if (file_type_list->types[p->entries[i].type]) {
             if (lstat(p->entries[i].dst, &stat_buf) == 0) {
                 add_file_to_list(NULL, p->entries[i].dst, l);
             }
@@ -714,22 +732,26 @@ static void find_existing_files(Package *p, FileList *l, uint64_t flag)
 /*
  * ignore_conflicting_file() - ignore (i.e., do not put it on the list
  * of files to backup) the conflicting file 'filename' if requiredString
- * is non-NULL and we cannot find the string in 'filename'.
+ * is non-NULL and we cannot find the string in 'filename', or if the
+ * file only conflicts on specific architectures, and the file's
+ * architecture does not match.
  */
 
 static int ignore_conflicting_file(Options *op,
                                    const char *filename,
-                                   const char *requiredString)
+                                   const ConflictingFileInfo info)
 {
     int fd = -1;
     struct stat stat_buf;
     char *file = MAP_FAILED;
     int ret = FALSE;
-    int i, len;
+    int i, len, size = 0;
 
-    /* if no requiredString, do not ignore this conflicting file */
+    /* if no requiredString, do not check for the required string */
 
-    if (!requiredString) return FALSE;
+    if (!info.requiredString) return ret;
+
+    ret = FALSE;
 
     if ((fd = open(filename, O_RDONLY)) == -1) {
         ui_error(op, "Unable to open '%s' for reading (%s)",
@@ -743,11 +765,13 @@ static int ignore_conflicting_file(Options *op,
         goto cleanup;
     }
 
-    if (!stat_buf.st_size) {
+    size = stat_buf.st_size;
+
+    if (!size) {
         goto cleanup;
     }
 
-    if ((file = mmap(0, stat_buf.st_size, PROT_READ,
+    if ((file = mmap(0, size, PROT_READ,
                      MAP_FILE | MAP_SHARED, fd, 0)) == MAP_FAILED) {
         ui_error(op, "Unable to map file '%s' for reading (%s)",
                  filename, strerror(errno));
@@ -763,11 +787,11 @@ static int ignore_conflicting_file(Options *op,
 
     ret = TRUE;
 
-    len = strlen(requiredString);
+    len = strlen(info.requiredString);
 
-    for (i = 0; (i + len) <= stat_buf.st_size; i++) {
-        if ((strncmp(&file[i], requiredString, len) == 0) &&
-            (((i + len) == stat_buf.st_size) || (file[i+len] == '\0'))) {
+    for (i = 0; (i + len) <= size; i++) {
+        if ((strncmp(&file[i], info.requiredString, len) == 0) &&
+            (((i + len) == size) || (file[i+len] == '\0'))) {
             ret = FALSE;
             break;
         }
@@ -778,7 +802,7 @@ static int ignore_conflicting_file(Options *op,
  cleanup:
 
     if (file != MAP_FAILED) {
-        munmap(file, stat_buf.st_size);
+        munmap(file, size);
     }
 
     if (fd != -1) {
@@ -801,7 +825,8 @@ static int ignore_conflicting_file(Options *op,
 static void find_conflicting_files(Options *op,
                                    char *path,
                                    ConflictingFileInfo *files,
-                                   FileList *l)
+                                   FileList *l,
+                                   const NoRecursionDirectory *skipdirs)
 {
     int i;
     char *paths[2];
@@ -819,9 +844,10 @@ static void find_conflicting_files(Options *op,
         case FTS_F:
         case FTS_SLNONE:
             for (i = 0; files[i].name; i++) {
+                /* end compare at len e.g. so "libGL." matches "libGL.so.1" */
                 if (!strncmp(ent->fts_name, files[i].name, files[i].len) &&
                     !ignore_conflicting_file(op, ent->fts_path,
-                                             files[i].requiredString)) {
+                                             files[i])) {
                     add_file_to_list(NULL, ent->fts_path, l);
                 }
             }
@@ -829,8 +855,17 @@ static void find_conflicting_files(Options *op,
 
         case FTS_DP:
         case FTS_D:
-            if (op->no_recursion)
+            if (op->no_recursion) {
                 fts_set(fts, ent, FTS_SKIP);
+            } else if (skipdirs) {
+                const NoRecursionDirectory *dir;
+                for (dir = skipdirs; dir->name; dir++) {
+                    if ((dir->level < 0 || dir->level >= ent->fts_level) &&
+                        strcmp(ent->fts_name, dir->name) == 0) {
+                        fts_set(fts, ent, FTS_SKIP);
+                    }
+                }
+            }
             break;
 
         default:
@@ -839,7 +874,7 @@ static void find_conflicting_files(Options *op,
              * and directories; traversing the hierarchy logically
              * to simplify handling of paths with symbolic links
              * to directories, we only need to handle broken links
-             * and, if recursion was disabled, directories.
+             * and, if recursion was not disabled, directories.
              */
             break;
         }
@@ -849,59 +884,69 @@ static void find_conflicting_files(Options *op,
 
 } /* find_conflicting_files() */
 
-
-
-
-/*
- * find_conflicting_libraries() - search for any conflicting
- * libraries in all relevant libdirs within the hierarchy under
- * the given prefix.
- */
-
-static void find_conflicting_libraries(Options *op,
-                                       const char *prefix,
-                                       ConflictingFileInfo *files,
-                                       FileList *l)
+void get_conflicting_file_info(const char *file, ConflictingFileInfo *cfi)
 {
-    int i, j;
-    char *paths[4];
+    char *c;
 
-    paths[0] = nvstrcat(prefix, "/", "lib", NULL);
-    paths[1] = nvstrcat(prefix, "/", "lib64", NULL);
-    paths[2] = nvstrcat(prefix, "/", "lib32", NULL);
-    paths[3] = NULL;
+    cfi->name = file;
 
-    for (i = 0; paths[i]; i++) {
-        for (j = 0; (j < 3) && paths[i]; j++) {
-            /*
-             * XXX Check if any one of the 'paths' entries really
-             * is a symbolic link pointing to one of the other
-             * entries. The logic could be made smarter, since it's
-             * unlikely that ../lib32 would be a symbolic link to
-             * ../lib64 or vice versa.
-             */
-            if (!paths[j] || (i == j)) continue;
+    /*
+     * match the names of DSOs with "libfoo.so*" by stopping any comparisons
+     * after ".so".
+     */
 
-            if (is_symbolic_link_to(paths[i], paths[j])) {
-                ui_expert(op, "The conflicting library search path "
-                          "'%s' is a symbolic link to the library "
-                          "search path '%s'; skipping '%s'.",
-                          paths[i], paths[j], paths[i]);
-                free(paths[i]); paths[i] = NULL;
-            }
-        }
-
-        if (paths[i]) find_conflicting_files(op, paths[i], files, l);
+    c = strstr(file, ".so.");
+    if (c) {
+        cfi->len = (c - file) + 3;
+    } else {
+        cfi->len = strlen(file);
     }
 
-    for (i = 0; i < 3; i++)
-        nvfree(paths[i]);
+    /*
+     * XXX avoid conflicting with libglx.so if it doesn't include the string
+     * "glxModuleData" to avoid removing the wrong libglx.so (bug 489316)
+     */
+
+    if (strncmp(file, "libglx.so", cfi->len) == 0) {
+        cfi->requiredString = "glxModuleData";
+    } else {
+        cfi->requiredString = NULL;
+    }
+}
+
+ConflictingFileInfo *build_conflicting_file_list(Options *op, Package *p)
+{
+    ConflictingFileInfo *cfList = NULL;
+    int index = 0;
+    int i;
+
+    // Allocate enough space for the whole file list, plus two extra files and
+    // a NULL at the end.
+    cfList = nvalloc((p->num_entries + 3) * sizeof(ConflictingFileInfo));
+
+    for (i = 0; i < p->num_entries; i++) {
+        PackageEntry *entry = &p->entries[i];
+        if (entry->caps.is_shared_lib && entry->caps.is_conflicting) {
+            get_conflicting_file_info(entry->name, &cfList[index++]);
+        }
+    }
+
+    /* XXX always conflict with these files if OpenGL files will be installed
+     * libGLwrapper.so: this library has an SONAME of libGL.so.1 (bug 74761) */
+    if (!op->no_opengl_files) {
+        get_conflicting_file_info("libGLwrapper.so", &cfList[index++]);
+    }
+
+    /* terminate the conflicting files list */
+    cfList[index].name = NULL;
+
+    return cfList;
+}
 
-} /* find_conflicting_libraries() */
 
 
 /*
- * condense_file_list() - Take a FileList stucture and delete any
+ * condense_file_list() - Take a FileList structure and delete any
  * duplicate entries in the list.  This is a pretty brain dead
  * brute-force algorithm.
  */
@@ -1031,6 +1076,7 @@ static void add_command(CommandList *c, int cmd, ...)
         c->cmds[n].s1 = nvstrdup(s);
         break;
       case DELETE_CMD:
+      case TOUCH_CMD:
         s = va_arg(ap, char *);
         c->cmds[n].s0 = nvstrdup(s);
         break;
